"""Microsoft Azure Document Intelligence API Handler"""

import io
import logging
import os
from dataclasses import dataclass, field

from azure.ai.documentintelligence import DocumentIntelligenceClient, AnalyzeDocumentLROPoller
from azure.ai.documentintelligence.models import AnalyzeResult, DocumentPage
from azure.core.credentials import AzureKeyCredential
from tqdm import tqdm

from . import layout
from ._config import PyPdfToTextConfig


logger = logging.getLogger(__name__)


@dataclass
class AzureDocIntelIntegrator:
    """
    Extract text from pdf images via calls to Azure Document Intelligence OCR API.
    """

    config: PyPdfToTextConfig = field(default_factory=PyPdfToTextConfig)
    client: DocumentIntelligenceClient | None = field(default=None, init=False, repr=False)
    last_result: AnalyzeResult = field(default_factory=lambda: AnalyzeResult({}), init=False)

    def create_client(self) -> bool:
        """
        Create an Azure DocumentIntelligenceClient based on current global
        constants and env var settings.

        The following may be set via env var prior to module import OR set via
        the corresponding self.config.<ENV_VARIABLE_NAME> global constant after
        module import.

        Constants/Environment Variables:
            AZURE_DOCINTEL_ENDPOINT: Azure Document Intelligence Instance Endpoint URL.
            AZURE_DOCINTEL_SUBSCRIPTION_KEY: Azure Document Intelligence Subscription Key.

        Returns:
            bool: True if client was created successfully. False otherwise.
        """
        endpoint = os.getenv("AZURE_DOCINTEL_ENDPOINT") or self.config.AZURE_DOCINTEL_ENDPOINT
        key = (
            os.getenv("AZURE_DOCINTEL_SUBSCRIPTION_KEY")
            or self.config.AZURE_DOCINTEL_SUBSCRIPTION_KEY
        )
        if endpoint and key:
            self.client = DocumentIntelligenceClient(endpoint, AzureKeyCredential(key))
            logger.info("Azure OCR Client Created: endpoint='%s'", endpoint)
            return True
        logger.error("Failed to create Azure OCR Client at endpoint='%s'", endpoint)
        return False

    def reset(self):
        """Clear last_result from previous run."""
        self.last_result = AnalyzeResult({})

    def ocr_pages(self, pdf: bytes, pages: list[int]) -> list[str]:
        """
        Read the text from supplied pdf page indices.

        Args:
            pdf: bytes of a pdf file
            pages: list of pdf page indices to OCR

        Returns:
            list[str]: list of strings containing structured text extracted
                from each supplied page index.
        """
        if self.config.AZURE_DOCINTEL_AUTO_CLIENT and self.client is None:
            self.create_client()
        if self.client is None:
            logger.error(
                "Azure OCR API not available. Did you create a client? Returning empty list."
            )
            return []
        assert self.client is not None
        logger.info("Sending pdf of %s bytes for OCR of %s pages.", len(pdf), len(pages))
        poller: AnalyzeDocumentLROPoller = self.client.begin_analyze_document(
            model_id="prebuilt-read",
            body=io.BytesIO(pdf),
            pages=",".join(str(pg + 1) for pg in pages),
        )
        self.last_result = poller.result(self.config.AZURE_DOCINTEL_TIMEOUT)
        logger.info("%s pages OCR'd successfully. Creating fixed width pages.", len(pages))
        ocr_pbar = tqdm(
            self.last_result.pages,
            desc="Processing OCR results...",
            disable=self.config.DISABLE_PROGRESS_BAR,
            position=self.config.PROGRESS_BAR_POSITION,
            leave=None,
        )
        results: list[str] = [
            layout.fixed_width_page(doc_page, self.config) for doc_page in ocr_pbar
        ]
        return results

    def handwritten_ratio(
        self,
        page_index: int,
        handwritten_confidence_limit: float | None = None,
    ) -> float:
        """
        Given a page *index*, returns the ratio of handwritten to total characters on the page.

        Args:
            page_index: the 0-based index of the page to analyze
            handwritten_confidence_limit: deprecated. use config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT

        Returns:
            float: 0.0 if the supplied page index was not OCR'd or has no text. Otherwise
            the ratio of the sum of all handwritten spans on the page to the total page span.
        """
        if handwritten_confidence_limit is not None:
            logger.warning(
                "Arg 'handwritten_confidence_limit' is no longer supported."
                " Supply the desired value via `self.config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT`."
                "\nrequested limit: %.2f (from arg)"
                "\neffective limit: %.2f (from self.config)",
                handwritten_confidence_limit,
                self.config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT,
            )

        if _selected_page := self.page_at_index(page_index):
            # a page should only have one span, but we'll treat as if there could be more
            # just in case. Get the min offset from all spans as the start and the max
            # offset + length as the page end.
            page_start = min(span.offset for span in _selected_page.spans)
            page_end = max(span.offset + span.length for span in _selected_page.spans)
            if page_end - page_start <= 0:
                # whoops! something's wrong. We should probably throw an exception here, but
                # we'll fail open for now as it fits our use case.
                logger.warning(
                    "Error calculating handwritten ratio for page at index %s:"
                    " page span start (%s) >= end (%s)",
                    page_index,
                    page_start,
                    page_end,
                )
                return 0.0
            # lets get the sum of span lengths for all is_handwritten styles with confidences
            # >= our threshold that also occur between page_start and page_end!
            handwritten_length = sum(
                (
                    (span.offset + min(span.length, page_end)) - span.offset
                    for style in (self.last_result.styles or [])
                    if style.is_handwritten
                    and style.confidence >= self.config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT
                    for span in style.spans
                    if page_start <= span.offset < page_end
                ),
                start=0,
            )
            # Guess we'll cap our value at 1.0. We should probably throw and exception here
            # also, but again we'll fail open for now as it suites our use case.
            ratio = handwritten_length / (page_end - page_start)
            if ratio > 1.0:
                logger.warning("Handwritten ratio of page index at %s capped at 1.0", page_index)
                return 1.0
            return ratio
        # page was not OCR'd return 0.0 default.
        return 0.0

    def rotation_degrees(self, page_index: int) -> float:
        """
        Given a page *index*, returns the degrees of rotation of the page reported by Azure.

        Args:
            page_index: the 0-based index of the page to analyze

        Returns:
            float: 0.0 if the supplied page index was not OCR'd. Otherwise
                the page's reported rotation in degrees.
        """
        if _selected_page := self.page_at_index(page_index):
            angle = _selected_page.angle or 0.0
            if abs(angle) > self.config.MIN_OCR_ROTATION_DEGREES:
                logger.debug("Page at index %s is rotated %.2f degrees", page_index, angle)
                return angle
        return 0.0

    def page_at_index(self, page_index: int) -> DocumentPage | None:
        """
        Returns the DocumentPage instance having the given page *index* or None.

        Args:
            page_index: the 0-based index of the page to analyze

        Returns:
            DocumentPage | None: None if the supplied page index was not OCR'd.
        """
        if any(
            # find the page at the supplied index and report its angle. otherwise return 0.0.
            (_selected_page := page).page_number == page_index + 1
            for page in self.last_result.pages
        ):
            return _selected_page
        # page was not OCR'd. Return None.
        return None


AZURE_READ = AzureDocIntelIntegrator()



"""Extract text from pdf pages from codebehind or Azure OCR as required"""

__version__ = "0.2.2"

import io
import logging
from pathlib import Path

from pypdf import PdfReader

from . import layout
from ._config import constants, PyPdfToTextConfig, PyPdfToTextConfigOverrides
from .azure_docintel_integrator import AZURE_READ
from .batch import PdfExtractBatch
from .pdf_extract import PdfExtract


logger = logging.getLogger(__name__)


def pdf_text_pages(
    pdf_reader: PdfReader | io.BytesIO | bytes,
    debug_path: Path | None = None,
    page_indices: list[int] | None = None,
    **kwargs,  # prevent errors due to bad args in upstream config dicts
) -> list[str]:
    """
    Extract text from PDF pages and return as a list of multiline strings.

    Uses PDF code-behind by default. Triggers Azure OCR if the fraction
    of pages having fewer than `MIN_LINES_OCR_TRIGGER` lines is greater than
    or equal to `TRIGGER_OCR_PAGE_RATIO`.

    NOTE: kwargs will accept lowercase variants of all `PyPdfToTextConfig` members.

    NOTE: *NOT* thread safe. Use `pdf_extract/PdfExtract` for thread safe operations.

    Args:
        pdf_reader (PdfReader | io.BytesIO | bytes): Pdf with pages to extract
            as an already instantiated PdfReader or the raw pdf bytes or BytesIO.
        debug_path (Path | None, optional): Path to write pypdf debug files to.
            Defaults to None.
        page_indices (list[int] | None): if provided, only extract text from
            the listed page indices. Default is None (extract all pages).

    Returns:
        list[str]: a string of text extracted for each page
    """
    config = PyPdfToTextConfig(
        overrides=PyPdfToTextConfigOverrides(
            **{
                _upper: kwargs.pop(k)
                for k in list(kwargs.keys())
                if hasattr(constants, _upper := k.upper())
            }
        ),
    )
    pdf_extract = PdfExtract(
        pdf_reader,
        config,
        debug_path=debug_path,
        azure=AZURE_READ,
    )
    page_indices = page_indices or list(range(len(pdf_extract.extracted_pages)))
    return [pg.text for idx, pg in enumerate(pdf_extract.extracted_pages) if idx in page_indices]


def pdf_text_page_lines(
    pdf_reader: PdfReader | io.BytesIO | bytes,
    debug_path: Path | None = None,
    page_indices: list[int] | None = None,
    **kwargs,  # prevent errors due to bad args in upstream config dicts
) -> list[list[str]]:
    """
    Extract text from PDF pages and return as a list of lines for each page.

    Uses PDF code-behind by default. Triggers Azure OCR if the fraction
    of pages having fewer than `MIN_LINES_OCR_TRIGGER` lines is greater than
    or equal to `TRIGGER_OCR_PAGE_RATIO`.

    NOTE: kwargs will accept lowercase variants of all `PyPdfToTextConfig` members.

    NOTE: *NOT* thread safe. Use `pdf_extract/PdfExtract` for thread safe operations.

    Args:
        pdf_reader (PdfReader | io.BytesIO | bytes): Pdf with pages to extract
            as an already instantiated PdfReader or the raw pdf bytes or BytesIO.
        debug_path (Path | None, optional): Path to write pypdf debug files to.
            Defaults to None.
        page_indices (list[int] | None): if provided, only extract text from
            the listed page indices. Default is None (extract all pages).

    Returns:
        list[list[str]]: a list of lines of text extracted for each page
    """
    return [
        pg.splitlines() for pg in pdf_text_pages(pdf_reader, debug_path, page_indices, **kwargs)
    ]


__all__ = [
    "constants",
    "layout",
    "AZURE_READ",
    "pdf_text_pages",
    "pdf_text_page_lines",
    "PyPdfToTextConfig",
    "PyPdfToTextConfigOverrides",
    "PdfExtract",
    "PdfExtractBatch",
]



"""PDF text extraction with OCR fallback and page manipulation utilities."""

from __future__ import annotations

import io
import json
import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, TYPE_CHECKING

from azure.ai.documentintelligence.models import DocumentPage
from pypdf import PdfReader, PdfWriter, PageObject
from pypdf.generic import DictionaryObject, NullObject
from tqdm import tqdm

from ._config import PyPdfToTextConfig, PyPdfToTextConfigOverrides
from .azure_docintel_integrator import AzureDocIntelIntegrator


try:
    import boto3
except ImportError:
    boto3 = None


try:
    from PIL import Image, ImageOps
except ImportError:
    Image = None
    ImageOps = None


if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client


logger = logging.getLogger(__name__)


@dataclass
class ExtractedPage:
    """
    Represents a single extracted page from a PDF with its metadata.

    This dataclass encapsulates all information about a page after text extraction,
    including the source of the text (embedded or OCR), handwritten content ratio,
    and references to the underlying page objects.

    Attributes:
        page_obj: The pypdf PageObject instance for this page.
        handwritten_ratio: Ratio of handwritten to total characters (0.0 to 1.0).
            Always 0.0 for embedded text pages.
        text: The extracted text content from the page.
        source: Indicates whether text was extracted from embedded PDF content
            ("embedded") or via OCR ("OCR").
        azure_page: The Azure DocumentPage instance if this page was OCR'd,
            None for embedded text pages.
    """

    page_obj: PageObject
    handwritten_ratio: float
    text: str
    source: Literal["embedded", "OCR"] = "embedded"
    azure_page: DocumentPage | None = None


class PdfExtract:
    """
    Turns PDF bytes into structured text pages with additional utilities for
    correcting page orientations, inspecting content quality, and producing child
    PDFs.

    Args:
        pdf (str | Path | bytes | io.BytesIO | PdfReader): if str, check for s3
            download if boto3 is installed and value is an s3 uri (i.e. prefixed 's3://').
            otherwise attempt read from disk.
        config: a PyPdfToTextConfig instance that inherits from the global base config
            `constants` by default. See PyPdfToTextConfig docstring for more info.

    KwArgs:
        debug_path (Path | None, optional): Path to write pypdf debug files to.
            Defaults to None.
        replace_byte_codes (dict[bytes, bytes] | None): if supplied, raw
            text is cast to bytes, the dict keys are replaced with values,
            and resulting bytes are cast back to text. Used to replace custom
            glyphs defined in PDFs w/ (roughly) equivalent unicode characters,
            e.g. 'anesthesia billing print set' checkbox handling.
        init_extracted_pages (list[ExtractedPage]): initializes the internal self._extracted_pages
            attribute to avoid re-extracting the same pages in child PDFs.
        compressed (bool): indicates that image compression has already been
            performed for the supplied Pdf.
    """

    def __init__(
        self,
        pdf: str | Path | bytes | io.BytesIO | PdfReader,
        config: PyPdfToTextConfig | None = None,
        **kwargs,
    ) -> None:
        self.config = config or PyPdfToTextConfig()
        self.corruption_detected: bool = False
        self.body: bytes
        self.debug_path: Path | None = kwargs.get("debug_path")
        self.compressed: bool = kwargs.get("compressed", False)
        self._extracted_pages: list[ExtractedPage] | None = kwargs.get("init_extracted_pages")
        self._azure: AzureDocIntelIntegrator | None = kwargs.get("azure")
        self._batch_mode: bool = kwargs.get("_batch_mode", False)
        self.ocr_page_idxs: list[int] = []
        self._reader: PdfReader | None = None
        self._writer: PdfWriter | None = None
        self._pbar: tqdm | None = None

        # set initial body value
        if isinstance(pdf, str):
            if pdf.startswith("s3://"):
                logger.info("Attempting to pull URI '%s' from s3", pdf)
                bucket, _, key = pdf[5:].partition("/")
                s3_object = self.s3.get_object(Bucket=bucket, Key=key)
                self.body = s3_object["Body"].read()
            else:
                self.body = Path(pdf).read_bytes()
        elif isinstance(pdf, Path):
            self.body = pdf.read_bytes()
        elif isinstance(pdf, bytes):
            self.body = pdf
        elif isinstance(pdf, io.BytesIO):
            self.body = pdf.getvalue()
        else:
            if not isinstance(pdf, PdfReader):
                raise TypeError(
                    f"`pdf` must be 'str | bytes | io.BytesIO | PdfReader': {type(pdf)=!r}"
                )
            self._reader = pdf
            assert isinstance(self._reader.stream, io.BytesIO)
            self.body = self._reader.stream.getvalue()

    @property
    def extracted_pages(self) -> list[ExtractedPage]:
        """
        A list of ExtractedPage objects containing text and metadata
        for each page in the source PDF.
        """
        if not self._extracted_pages:
            self._extracted_pages = self._extract_pages()
        return self._extracted_pages

    @property
    def text_pages(self) -> list[str]:
        """
        A list of multiline strings containing a structured text representation
        of each page in the source PDF.
        """
        return [ext_pg.text for ext_pg in self.extracted_pages]

    @property
    def text_page_lines(self) -> list[list[str]]:
        """
        A list of lists of strings with each sublist containing the lines
        in the structured text representation of each page from the source PDF.
        """
        return [ext_pg.text.splitlines() for ext_pg in self.extracted_pages]

    @property
    def text(self) -> str:
        """Text extracted from all pages in a single string."""
        return "\n".join(ext_pg.text for ext_pg in self.extracted_pages)

    @property
    def reader(self) -> PdfReader:
        """The PDF reader used for text extraction. **NOT THREAD SAFE.**"""
        if self._reader is None:
            logger.debug("Initializing PdfReader for PdfExtract")
            self._reader = PdfReader(io.BytesIO(self.body))
        return self._reader

    @property
    def writer(self) -> PdfWriter:
        """The PDF writer used for image compression, child creation,
        and page rotations. **NOT THREAD SAFE.**"""
        if self._writer is None:
            logger.debug("Initializing PdfWriter for PdfExtract")
            if self._extracted_pages:
                self._writer = PdfWriter()
                self._writer.append(
                    self.reader, pages=[ext_pg.page_obj for ext_pg in self._extracted_pages]
                )
            else:
                self._writer = PdfWriter(clone_from=self.reader)
        return self._writer

    @property
    def s3(self) -> "S3Client":
        """S3 client for this instance. Raises ImportError if boto3 is not installed."""
        if boto3 is None:
            raise ImportError('boto3 not found. Run `pip install pypdftotext["s3"]`.')
        if not hasattr(self, "_s3"):
            # lazy load this one...
            logger.debug("Initializing boto3 s3 client for PdfExtract")
            self._s3 = boto3.client(  # pylint: disable=attribute-defined-outside-init
                service_name="s3",
                aws_access_key_id=self.config.AWS_ACCESS_KEY_ID,
                aws_secret_access_key=self.config.AWS_SECRET_ACCESS_KEY,
                aws_session_token=self.config.AWS_SESSION_TOKEN,
            )
        return self._s3

    def _embedded_text(self, pg: PageObject, pg_idx: int) -> str | int:
        """Get embedded PDF text using pypdf's layout mode"""
        if self.corruption_detected or self.config.SUPPRESS_EMBEDDED_TEXT:
            txt = ""
        else:
            try:
                txt = pg.extract_text(
                    extraction_mode="layout",
                    layout_mode_space_vertically=self.config.PRESERVE_VERTICAL_WHITESPACE,
                    layout_mode_scale_weight=self.config.SCALE_WEIGHT,
                    layout_mode_debug_path=self.debug_path,
                    layout_mode_font_height_weight=self.config.FONT_HEIGHT_WEIGHT,
                )
            except (ZeroDivisionError, TypeError):
                txt = "\n".join(
                    line
                    for line in pg.extract_text().splitlines()
                    if line.strip() or self.config.PRESERVE_VERTICAL_WHITESPACE
                )
        if len(txt) > self.config.MAX_CHARS_PER_PDF_PAGE:
            self.corruption_detected = True
            if self._pbar:
                self._pbar.set_postfix_str("!!! CORRUPTION DETECTED !!!")
            logger.warning(
                "Clearing corrupt pdf text pg_idx=%s; len(txt)=%s > %s char limit.",
                pg_idx,
                len(txt),
                self.config.MAX_CHARS_PER_PDF_PAGE,
            )
            txt = ""
        # this originally compared `len(txt.splitlines())` which was
        # VERY inefficient. The + 1 below preserves the original
        # behavior for the `min_lines_ocr_trigger` parameter
        line_count_less_than_ocr_trigger = txt.count("\n") + 1 <= self.config.MIN_LINES_OCR_TRIGGER
        # add as an OCR candidate if page has too few lines.
        if not self.config.DISABLE_OCR and line_count_less_than_ocr_trigger:
            return pg_idx
        return txt

    def _extract_pages(self) -> list[ExtractedPage]:
        """
        Extract text from PDF pages and return as a list of ExtractedPage objects.

        Uses PDF code-behind by default. Triggers Azure OCR per config if the fraction
        of pages having fewer than `MIN_LINES_OCR_TRIGGER` lines is greater than
        or equal to `TRIGGER_OCR_PAGE_RATIO`.

        Returns:
            list[ExtractedPage]: a list of extracted pages
        """

        self._pbar = tqdm(
            enumerate(self.reader.pages),
            desc="Extracting text",
            disable=self.config.DISABLE_PROGRESS_BAR,
            position=self.config.PROGRESS_BAR_POSITION,
            leave=None,
        )
        pre_ocr = [self._embedded_text(page, page_index) for page_index, page in self._pbar]
        self._extracted_pages = [
            ExtractedPage(pg, 0.0, txt if isinstance(txt, str) else "")
            for txt, pg in zip(pre_ocr, self.reader.pages)
        ]
        self.ocr_page_idxs = [itm for itm in pre_ocr if isinstance(itm, int)]
        if self._azure:
            self._azure.config = self.config
        azure = AzureDocIntelIntegrator(self.config) if self._azure is None else self._azure

        # Parallel Azure OCR API calls will be made later if in batch mode.
        if not self._batch_mode:
            self.ocr(azure)

        # perform byte code substitutions per 'replace_byte_codes' arg
        if self.config.REPLACE_BYTE_CODES:
            replacements = [
                (old_bytes.decode(), new_bytes.decode())
                for old_bytes, new_bytes in self.config.REPLACE_BYTE_CODES.items()
            ]
            for ext_pg in self._extracted_pages:
                if not ext_pg.text:
                    continue
                for old_, new_ in replacements:
                    ext_pg.text = ext_pg.text.replace(old_, new_)

        logger.info("Text extraction complete.")
        return self._extracted_pages

    def ocr(self, azure: AzureDocIntelIntegrator):
        """
        Run OCR on identified indices if the fraction of pages having fewer
        than `self.config.MIN_LINES_OCR_TRIGGER` lines is greater than or equal to
        `self.config.TRIGGER_OCR_PAGE_RATIO`. Updates the proper self.extracted_pages
        entries with OCR'd text and handwritten ratios if OCR is triggered.
        """
        rotated_pages = False  # track whether we rotated any pages and regenerate self.body if so.
        # do not OCR unless the number of pages requiring OCR / total pages exceeds a target ratio.
        # Skip OCR if in batch mode (will be handled by batch processor)
        if (
            len(self.ocr_page_idxs) / len(self.extracted_pages)
            >= self.config.TRIGGER_OCR_PAGE_RATIO
        ):
            # if not in batch mode, replacements are handled globally in _extract_pages.
            if not self._batch_mode:
                replacements = []
            else:
                replacements = [
                    (old_bytes.decode(), new_bytes.decode())
                    for old_bytes, new_bytes in (self.config.REPLACE_BYTE_CODES or {}).items()
                ]

            ocr_pages = azure.ocr_pages(self.body, self.ocr_page_idxs)
            if self.debug_path:
                (self.debug_path / "ocr_pages.json").write_text(
                    json.dumps(ocr_pages, indent=2, default=str), "utf-8"
                )
            for ocr_idx, og_pg_idx in enumerate(self.ocr_page_idxs):
                ext_pg = self.extracted_pages[og_pg_idx]
                txt = ocr_pages[ocr_idx]
                if len(txt) > self.config.MAX_CHARS_PER_PDF_PAGE:
                    logger.warning(
                        "Clearing corrupt OCR text pg_idx=%s; len(txt)=%s > %s char limit."
                        " Does page contain multiple text orientations?",
                        og_pg_idx,
                        len(txt),
                        self.config.MAX_CHARS_PER_PDF_PAGE,
                    )
                    txt = ""
                elif rotation := azure.rotation_degrees(og_pg_idx):
                    # rotations can only be applied to pages in 90 degree increments.
                    # do not report rotated content if applied rotation is 0.
                    if applied_rotation := -90 * int(round(rotation / 90.0)):
                        rotated_pages = True
                        ext_pg.page_obj.rotation += applied_rotation

                # perform byte code substitutions per 'replace_byte_codes' arg if in batch mode
                if replacements and txt:
                    for old_, new_ in replacements:
                        txt = txt.replace(old_, new_)

                ext_pg.text = txt
                ext_pg.source = "OCR"
                ext_pg.handwritten_ratio = azure.handwritten_ratio(og_pg_idx)
                ext_pg.azure_page = azure.page_at_index(og_pg_idx)

        if rotated_pages:
            logger.debug("Regenerating PdfExtract body with corrected page orientations.")
            self._regenerate_body()

    def child(
        self,
        page_indices: list[int] | tuple[int, int] | None = None,
        config_overrides: PyPdfToTextConfigOverrides | None = None,
    ) -> PdfExtract:
        """
        Creates a child PdfExtract instance for the selected pages preserving
        extracted text, image compressions, and page reorientations from the parent.

        Args:
            page_indices (list[int] | tuple[int, int] | None): a list of 0-based page indices
                OR a tuple of start page index, stop page index (inclusive) to include. If
                None (default), all pages are included in the child.
            config_overrides (PyPdfToTextConfigOverrides | None): settings to override in the
                child instance.

        Returns:
            PdfExtract: a child instance containing the pages specified.
        """
        if isinstance(page_indices, tuple):
            page_indices = list(range(page_indices[0], page_indices[1] + 1))
        elif page_indices is None:
            page_indices = list(range(len(self.reader.pages)))
        return PdfExtract(
            pdf=self.clip_pages(page_indices),
            config=PyPdfToTextConfig(base=self.config, overrides=config_overrides),
            init_extracted_pages=[
                pg for idx, pg in enumerate(self.extracted_pages) if idx in page_indices
            ],
            compressed=self.compressed,
        )

    def compress_images(
        self,
        white_point: int = 220,
        max_overscale: float = 2,
        aspect_tolerance: float = 1e-3,
        force: bool = False,
    ):
        """
        Reduces the size of the pdf by converting all images to grey scale and
        downsampling full page images more than 2x larger than the displayed
        area of the PDF.

        Args:
            white_point: pixel values greater than this are set to white (255) after
                casting to grey scale. Reduces noise for a sharper image. Values >= 256
                effectively disable denoising.
            max_overscale: the width of the image must be > max_overscale * page width
                for downsampling. the scale factor of the downsampled image is set to
                (max_overscale * page width) / image width.
            aspect_tolerance: an image is considered to be a full page image and
                eligible for downsampling if and only if the aspect ratio (width/height)
                of the image is within this tolerance of the aspect ratio of the entire page.
            force (bool): if true, ignore `self.compressed` and re-run the algorithm.

        Returns:
            None: images are updated in place in the PDF.
        """
        if Image is None or ImageOps is None:
            raise ImportError("PIL not found. Run `pip install pypdftotext[image]`.")
        if self.compressed and not force:
            logger.info("PdfExtract images are already compressed. No action taken.")
            return  # we've already compressed these images
        for ip, page in enumerate(self.writer.pages):
            for ii, img in enumerate(page.images):
                if not isinstance(img.image, Image.Image):
                    logger.debug("Bad image: page index %s image index %s", ip, ii)
                    continue
                new_img = img.image.convert("L").point(
                    lambda x: x if x < white_point else 256  # type: ignore[reportOperatorIssue]
                )
                page_aspect = abs(page.mediabox.width / page.mediabox.height)
                img_aspect = abs(new_img.width / new_img.height)
                if (
                    new_img.width > max_overscale * page.mediabox.width
                    and abs(page_aspect - img_aspect) < aspect_tolerance
                ):
                    factor = (max_overscale * page.mediabox.width) / new_img.width
                    logger.debug("Scaling pg idx %s img idx %s with factor=%.2f", ip, ii, factor)
                    new_img = ImageOps.scale(new_img, factor, resample=Image.Resampling.LANCZOS)
                img.replace(new_img, resolution=300)
        self._regenerate_body()
        self.compressed = True

    def _regenerate_body(self):
        new_body_io = io.BytesIO()
        self.writer.write(new_body_io)
        self.body = new_body_io.getvalue()
        # force reader/writer properties to reinitialize from self.body
        self._reader = None
        self._writer = None

    def clip_pages(self, page_indices: list[int] | tuple[int, int]) -> bytes:
        """
        Clip specific pages from a source pdf into a new pdf document, removing any
        globally defined images that do not appear in the clipped pages. If text
        extraction has already occurred, page orientations will also be corrected.

        Args:
            page_indices (list[int] | tuple[int, int]): a list of 0-based page indices
                OR a tuple of start page index, stop page index (inclusive) to include.

        Returns:
            bytes: the new pdf
        """
        if isinstance(page_indices, tuple):
            page_indices = list(range(page_indices[0], page_indices[1] + 1))
        pdf_writer = PdfWriter()
        pdf_writer.append(self.reader, pages=page_indices)
        ootb_bytesio = io.BytesIO()
        pdf_writer.write(ootb_bytesio)
        # Optimize storage size and prevent possible 'bleed through' of xobject data.
        xobjs: list[DictionaryObject] = []
        xobj_pages: list[list[int]] = []

        for i, out_page in enumerate(pdf_writer.pages):
            # collect all /Resources XObject references and the list of page indices
            # that reference each of them
            rsrcs = out_page.get_inherited("/Resources", {})
            if "/XObject" in rsrcs:
                xobj = rsrcs["/XObject"]
                if xobj in xobjs:
                    xobj_pages[xobjs.index(xobj)].append(i)
                else:
                    xobjs.append(xobj)
                    xobj_pages.append([i])

        replaced_one = False  # don't waste time rerunning write operation if no xobjs are cleared

        for page_group, xobj in zip(xobj_pages, xobjs):
            # find the names of XObjects that are actually referenced by a page in the output.
            name_regex = re.compile(r"(" + r"|".join(xobj.keys()) + r")\s", re.MULTILINE)
            refd_names = set(
                _m
                for i in page_group
                if (_contents := pdf_writer.pages[i].get_contents()) is not None
                for _m in name_regex.findall(str(_contents.get_data(), "utf-8", "ignore"))
            )
            # add the names of any XObjects referenced by the XObjects captured via regex.
            refd_names.update(
                xobj[img]["/SMask"]["/Name"]  # pyright: ignore
                for img in refd_names.copy()
                if img in xobj
                and "/SMask" in xobj[img]  # pyright: ignore
                and "/Name" in xobj[img]["/SMask"]  # pyright: ignore
            )
            logger.debug(
                "page_group='%s' references refd_names=%s; clearing: %s",
                page_group,
                refd_names,
                xobj.keys() - refd_names,
            )
            for unrefd_img in xobj.keys() - refd_names:
                replaced_one = True
                pdf_writer._replace_object(  # pylint:disable=protected-access
                    xobj[unrefd_img].indirect_reference, NullObject()  # pyright: ignore
                )

        # if we nulled out any xobjs, regenerate output bytes.
        if replaced_one:
            optimized_bytesio = io.BytesIO()
            pdf_writer.write(optimized_bytesio)
            return optimized_bytesio.getvalue()

        # otherwise, return ootb bytes.
        return ootb_bytesio.getvalue()

    def handwritten_ratio(
        self,
        page_index: int,
        handwritten_confidence_limit: float | None = None,
    ) -> float:
        """
        Given a page *index*, returns the ratio of handwritten to total characters on the page.

        Args:
            page_index: the 0-based index of the page to analyze
            handwritten_confidence_limit: deprecated. use config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT

        Returns:
            float: 0.0 if the supplied page index was not OCR'd or has no text. Otherwise
            the ratio of the sum of all handwritten spans on the page to the total page span.
        """
        if handwritten_confidence_limit is not None:
            logger.warning(
                "Arg 'handwritten_confidence_limit' is no longer supported."
                " Supply the desired value via `self.config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT`."
                "\nrequested limit: %.2f (from arg)"
                "\neffective limit: %.2f (from self.config)",
                handwritten_confidence_limit,
                self.config.OCR_HANDWRITTEN_CONFIDENCE_LIMIT,
            )
        return self.extracted_pages[page_index].handwritten_ratio



"""Batch processing module for efficient parallel OCR of multiple PDFs."""

from __future__ import annotations

import io
import logging
from collections.abc import Mapping, Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from pathlib import Path

from azure.core.exceptions import AzureError
from pypdf import PdfReader
from tqdm import tqdm

from ._config import PyPdfToTextConfig
from .azure_docintel_integrator import AzureDocIntelIntegrator
from .pdf_extract import PdfExtract


logger = logging.getLogger(__name__)


class PdfExtractBatch:
    """
    Processes multiple PDFs efficiently with sequential embedded text extraction
    and parallel OCR processing.

    This class maintains full backward compatibility while enabling efficient
    batch processing of multiple PDFs. Embedded text extraction occurs sequentially
    for all PDFs, then all pages needing OCR are submitted in a single batch
    for parallel processing.

    Args:
        pdfs: List or mapping of PDF inputs (str | Path | bytes | io.BytesIO | PdfReader)
        config: Configuration to use for all PDFs (defaults to PyPdfToTextConfig())
        **kwargs: Additional arguments passed to PdfExtract instances

    Usage:
        pdfs = ["file1.pdf", "file2.pdf", Path("file3.pdf")]
        batch = PdfExtractBatch(pdfs)
        pdf_extracts = batch.extract_all()
        # pdf_extracts is a list of PdfExtract objects with text already extracted
    """

    def __init__(
        self,
        pdfs: (
            Sequence[str | Path | bytes | io.BytesIO | PdfReader]
            | Mapping[str, str | Path | bytes | io.BytesIO | PdfReader]
        ),
        config: PyPdfToTextConfig | None = None,
        **kwargs,
    ) -> None:
        if not isinstance(pdfs, (list, dict)):
            raise TypeError(
                f"PdfExtractBatch input should be list or dict, received {type(pdfs)=!r}"
            )
        self.pdfs = (
            pdfs if isinstance(pdfs, dict) else {f"PDF[{i}]": pdf for i, pdf in enumerate(pdfs)}
        )
        self.config = config or PyPdfToTextConfig()
        self.kwargs = kwargs
        logger.info("Starting batch extraction for %s PDFs", len(self.pdfs))
        # Create the pdf extract objects but don't extract text until 'process' is called.
        self.pdf_extracts, self.s3_errors = self._pull_s3_parallel()

    def _pull_s3_parallel(self) -> tuple[dict[str, PdfExtract], dict[str, str]]:
        """Parallelize calls to s3 if present. Returns a dict of extracts that were created
        successfully and a dict of pdf name: s3 uri for failures"""
        s3_uris = {
            k: v for k, v in self.pdfs.items() if isinstance(v, str) and v.startswith("s3://")
        }
        pdf_extracts: dict[str, PdfExtract] = {
            pdf_name: PdfExtract(
                pdf=pdf, config=self.config, **{**self.kwargs, "_batch_mode": True}
            )
            for pdf_name, pdf in self.pdfs.items()
            if pdf_name not in s3_uris or len(s3_uris) == 1
        }
        s3_errors = {}
        if len(s3_uris) <= 1:
            return pdf_extracts, {}
        with ThreadPoolExecutor(
            max_workers=min(len(s3_uris), self.config.MAX_WORKERS)
        ) as executor:

            futures: list[Future[tuple[str, PdfExtract | Exception]]] = []
            for pdf_name, s3_uri in s3_uris.items():
                futures.append(executor.submit(self._extract_from_s3_uri, (pdf_name, s3_uri)))

            # Process results as they complete
            pbar = tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Downloading Objects from S3",
                disable=self.config.DISABLE_PROGRESS_BAR,
                position=self.config.PROGRESS_BAR_POSITION,
                leave=None,
            )

            for i, future in enumerate(pbar):
                pdf_name, extract_or_error = future.result()
                logger.debug("S3 Download Complete: %r (%s/%s)", pdf_name, i, len(s3_uris))
                if isinstance(extract_or_error, Exception):
                    s3_errors[pdf_name] = extract_or_error
                else:
                    pdf_extracts[pdf_name] = extract_or_error
        return pdf_extracts, s3_errors

    def _extract_from_s3_uri(
        self, s3_uri_tuple: tuple[str, str]
    ) -> tuple[str, PdfExtract | Exception]:
        pdf_name, s3_uri = s3_uri_tuple
        try:
            extract = PdfExtract(
                pdf=s3_uri, config=self.config, **{**self.kwargs, "_batch_mode": True}
            )
            return pdf_name, extract
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(
                "S3 Download Error: %r failed, %s",
                pdf_name,
                e,
                exc_info=logger.getEffectiveLevel() == logging.DEBUG,
            )
            return pdf_name, e

    def extract_all(self) -> dict[str, PdfExtract]:
        """Extract embedded text serially, then perform OCR operations in parallel."""
        # Step 1: Perform embedded text extraction
        self.pdf_extracts = self._extract_embedded_text()

        try:
            # Step 2: Perform batch OCR if needed
            self._perform_batch_ocr()
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(
                "PdfExtractBatch OCR Error: %e",
                e,
                exc_info=logger.getEffectiveLevel() == logging.DEBUG,
            )

        logger.info("Batch extraction complete for %s PDFs", len(self.pdf_extracts))
        return self.pdf_extracts

    def _extract_embedded_text(self) -> dict[str, PdfExtract]:
        """Create PdfExtract instances with embedded text extraction only."""
        pbar = tqdm(
            self.pdf_extracts.items(),
            desc="Extracting embedded text",
            disable=self.config.DISABLE_PROGRESS_BAR,
            position=self.config.PROGRESS_BAR_POSITION,
            leave=None,
        )

        for i, (pdf_name, pdf) in enumerate(pbar):
            logger.debug("Extracting text from %s (%s/%s)", pdf_name, i, len(self.pdfs))
            # Create PdfExtract with batch mode flag to prevent individual OCR
            pbar.set_postfix_str(pdf_name)
            _ = pdf.extracted_pages

        return self.pdf_extracts

    def _perform_batch_ocr(self) -> dict[str, PdfExtract]:
        """Perform OCR for all collected pages in parallel."""

        ocr_pdfs = {  # get pdfs that need OCR
            pdf_name: extract
            for pdf_name, extract in self.pdf_extracts.items()
            if (
                len(extract.ocr_page_idxs) / len(extract.extracted_pages)
                >= self.config.TRIGGER_OCR_PAGE_RATIO
            )
        }
        if not ocr_pdfs:
            logger.debug(
                "No PDFs met OCR criteria (MIN_LINES_OCR_TRIGGER=%s, TRIGGER_OCR_PAGE_RATIO=%s)",
                self.config.MIN_LINES_OCR_TRIGGER,
                self.config.TRIGGER_OCR_PAGE_RATIO,
            )
            return self.pdf_extracts
        total_pages = sum(len(ext.ocr_page_idxs) for ext in ocr_pdfs.values())
        logger.info("Submitting %s pages across %s PDFs for batch OCR", total_pages, len(ocr_pdfs))

        # Process PDFs in parallel using ThreadPoolExecutor
        with ThreadPoolExecutor(
            max_workers=min(len(ocr_pdfs), self.config.MAX_WORKERS)
        ) as executor:

            futures: list[Future[tuple[str, PdfExtract]]] = []
            for pdf_name, pdf_ext in ocr_pdfs.items():
                futures.append(executor.submit(self._ocr_single_pdf, (pdf_name, pdf_ext)))

            # Process results as they complete
            pbar = tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Processing OCR results",
                disable=self.config.DISABLE_PROGRESS_BAR,
                position=self.config.PROGRESS_BAR_POSITION,
                leave=None,
            )

            for i, future in enumerate(pbar):
                pdf_name, _ = future.result()
                logger.debug("OCR complete for %s (%s/%s)", pdf_name, i, len(ocr_pdfs))
        return self.pdf_extracts

    def _ocr_single_pdf(self, pdf_info: tuple[str, PdfExtract]) -> tuple[str, PdfExtract]:
        """OCR a single PDF's pages."""
        pdf_name, pdf_extract = pdf_info
        try:
            logger.debug("Begin OCR for %s", pdf_name)
            azure = AzureDocIntelIntegrator(self.config)
            if self.config.AZURE_DOCINTEL_AUTO_CLIENT and azure.client is None:
                azure.create_client()

            # Run OCR for this extract.
            pdf_extract.ocr(azure)
        except AzureError as azure_error:
            logger.error(
                "PdfExtractBatch Azure Error for %s: %s",
                pdf_name,
                azure_error,
                exc_info=logger.getEffectiveLevel() == logging.DEBUG,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(
                "PdfExtractBatch Error for %s: %s",
                pdf_name,
                e,
                exc_info=logger.getEffectiveLevel() == logging.DEBUG,
            )

        return pdf_name, pdf_extract



"""Functions for converting bbox info to a fixed width page layout"""

import logging
import math
from itertools import groupby

from azure.ai.documentintelligence.models import DocumentLine, DocumentPage

from ._config import constants, PyPdfToTextConfig

logger = logging.getLogger(__name__)


def rotated_bbox(
    line: DocumentLine,
    page: DocumentPage,
    min_ocr_rotation_degrees: float = constants.MIN_OCR_ROTATION_DEGREES,
) -> list[float]:
    """
    Rotate the bounding box for the line according to the angle of rotation
    reported for the page.

    Args:
        line: an Azure DocumentLine instance
        page: an Azure DocumentPage instance
        min_ocr_rotation_degrees: min reported rotation to apply.

    Returns:
        list[float]: list of 8 floats corresponding to the coordinates of the
            corners of the bounded region.

    Examples:
        >>> from azure.ai.documentintelligence.models import DocumentLine, DocumentPage
        >>> # Create a line with a bounding box at corners (1,2), (3,2), (3,3), (1,3)
        >>> line = DocumentLine(polygon=[1.0, 2.0, 3.0, 2.0, 3.0, 3.0, 1.0, 3.0])
        >>> # Page with no rotation
        >>> page = DocumentPage(page_number=1, angle=0.0, width=8.5, height=11.0)
        >>> rotated_bbox(line, page, min_ocr_rotation_degrees=5.0)
        [1.0, 2.0, 3.0, 2.0, 3.0, 3.0, 1.0, 3.0]
        >>> # Page with small rotation (below threshold)
        >>> page_small = DocumentPage(page_number=1, angle=3.0, width=8.5, height=11.0)
        >>> rotated_bbox(line, page_small, min_ocr_rotation_degrees=5.0)
        [1.0, 2.0, 3.0, 2.0, 3.0, 3.0, 1.0, 3.0]
        >>> # Page rotated 90 degrees (simulating a page that needs correction)
        >>> page_rotated = DocumentPage(page_number=1, angle=90.0, width=8.5, height=11.0)
        >>> result = rotated_bbox(line, page_rotated, min_ocr_rotation_degrees=5.0)
        >>> # Verify rotation occurred (values will differ from original)
        >>> result != [1.0, 2.0, 3.0, 2.0, 3.0, 3.0, 1.0, 3.0]
        True
    """
    if line.polygon is None:
        raise ValueError(f"Bad Azure DocumentLine: polygon is None for {line=}")
    if page.angle is None or abs(page.angle) < min_ocr_rotation_degrees:
        return line.polygon
    # We're *reversing* the reported angle, so convert as `-angle`
    angle = math.radians(-page.angle)
    _sin = math.sin(angle)
    _cos = math.cos(angle)
    height = page.height or 11
    width = page.width or 8.5
    p = line.polygon

    def _rotate_point(x: float, y: float) -> tuple[float, float]:
        return (
            _cos * (x - width / 2) - _sin * (y - height / 2) + width / 2,
            _sin * (x - width / 2) + _cos * (y - height / 2) + height / 2,
        )

    return [
        *_rotate_point(p[0], p[1]),
        *_rotate_point(p[2], p[3]),
        *_rotate_point(p[4], p[5]),
        *_rotate_point(p[6], p[7]),
    ]


class CharGroup:
    """
    Describes the font size and bounding box for a rendered text element.
    If multiple CharGroup instances fall on the same line, the text will be
    combined with proper spacing added according to x coordinate offsets
    during the `y_coordinate_groups` operation.

    NOTE: This class scales coordinates using OCR_POSITIONING_SCALE (default: 100)
    for x/y positioning and OCR_LINE_HEIGHT_SCALE (default: 50) for height
    calculations. The different scaling factors help prevent improper line
    splitting when pages have rotation angles.

    Args:
        line: an Azure DocumentLine instance
        page: an Azure DocumentPage instance
        config: a PyPdfToTextConfig instance inherits from global base config
            `constants` by default. See PyPdfToTextConfig docstring for more info.

    Attributes:
        tx (float): x coordinate of first character in CharGroup
        ty (float): y coordinate of first character in CharGroup
        effective_height (float): effective bbox height
        text (str): rendered text
        displaced_tx (float): x coordinate of the right edge of the CharGroup's bounding box
    """

    def __init__(
        self,
        line: DocumentLine,
        page: DocumentPage,
        config: PyPdfToTextConfig | None = None,
    ) -> None:
        config = config or PyPdfToTextConfig()
        if line.polygon is None:
            raise ValueError(f"Bad Azure DocumentLine: {line.polygon=!r}")
        bbox = rotated_bbox(line, page, config.MIN_OCR_ROTATION_DEGREES)
        self.tx: float = bbox[0] * config.OCR_POSITIONING_SCALE
        self.ty: float = bbox[1] * config.OCR_POSITIONING_SCALE
        self.effective_height: float = (bbox[-1] - bbox[1]) * config.OCR_LINE_HEIGHT_SCALE
        self.text: str = line.content
        self.displaced_tx: float = bbox[2] * config.OCR_POSITIONING_SCALE

    def offset_x_coords(self, offset: float) -> "CharGroup":
        """Decrement self.tx and self.displaced_tx by the offset and return self."""
        self.tx -= offset
        self.displaced_tx -= offset
        return self

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}:\n"
            f"  tx:               {self.tx:.2f}\n"
            f"  ty:               {self.ty:.2f}\n"
            f"  effective_height: {self.effective_height:.2f}\n"
            f"  text:             {self.text!r}\n"
            f"  displaced_tx:     {self.displaced_tx:.2f}"
        )


def dedented_groups(groups: list[CharGroup]) -> list[CharGroup]:
    """Finds the minimum x coord reported across all groups and decrements the
    x coords of all groups by this minimum. Ensures the rendered page output
    is aligned such that the leftmost rendered text appears in the first column
    of the page output.

    Returns:
        list[CharGroup]: sorted by y-coordinate (top to bottom) then x-coordinate (right to
            left), with x-coordinates adjusted to align leftmost text to position 0
    """
    min_x = min((x.tx for x in groups), default=0.0)
    dedented = [
        grp.offset_x_coords(min_x)
        # need to think about this sorting. Seems odd to negate
        # both and then reverse.
        for grp in sorted(groups, key=lambda x: (-x.ty, -x.tx), reverse=True)
        if grp.text
    ]
    if logger.getEffectiveLevel() == logging.DEBUG:
        logger.debug("**** DEDENTED GROUPS ****\n%s\n", "\n".join(str(grp) for grp in dedented))
    return dedented


def y_coordinate_groups(groups: list[CharGroup]) -> dict[int, list[CharGroup]]:
    """
    Group text operations by rendered y coordinate, i.e. the line number.

    Args:
        groups: list of CharGroup instances as returned by dedented_groups()

    Returns:
        dict[int, list[CharGroup]]: dict mapping y-coordinate positions to lists of
            CharGroup instances that appear on the same line
    """
    ty_groups = {
        ty: sorted(grp, key=lambda x: x.tx)
        for ty, grp in groupby(groups, key=lambda _grp: int(_grp.ty))
    }
    # combine groups whose y coordinates differ by less than the effective height
    # (accounts for mixed fonts and other minor oddities)
    last_ty = next(iter(ty_groups))
    last_txs = {int(_t.tx) for _t in ty_groups[last_ty] if _t.text.strip()}
    for ty in list(ty_groups)[1:]:
        fsz = min(ty_groups[_y][0].effective_height for _y in (ty, last_ty))
        txs = {int(_t.tx) for _t in ty_groups[ty] if _t.text.strip()}
        # prevent merge if both groups are rendering in the same x position.
        no_text_overlap = not txs & last_txs
        offset_less_than_font_height = abs(ty - last_ty) < fsz
        if no_text_overlap and offset_less_than_font_height:
            ty_groups[last_ty] = sorted(ty_groups.pop(ty) + ty_groups[last_ty], key=lambda x: x.tx)
            last_txs |= txs
        else:
            last_ty = ty
            last_txs = txs
    if logger.getEffectiveLevel() == logging.DEBUG:
        logger.debug(
            "**** Y COORDINATE GROUPS ****\n----- %s\n",
            "\n----- ".join(
                f"Y-Position: {ty} -----\n" + "\n".join(str(grp) for grp in grps)
                for ty, grps in ty_groups.items()
            ),
        )
    return ty_groups


def fixed_char_width(groups: list[CharGroup], scale_weight: float = 1.25) -> float:
    """
    Calculate average character width weighted by the length of the rendered
    text in each sample for conversion to fixed-width layout.

    Args:
        groups (list[CharGroup]): list of CharGroup instances created from
            the lines of an Azure OCR API response.
        scale_weight (float, optional): Weight factor applied to text length when
            calculating average character width. Default is 1.25.

    Returns:
        float: fixed character width

    """
    char_widths = []
    for _group in groups:
        _len = len(_group.text) * scale_weight
        char_widths.append(((_group.displaced_tx - _group.tx) / _len, _len))
    return sum(_w * _l for _w, _l in char_widths) / sum(_l for _, _l in char_widths)


def fixed_width_page(page: DocumentPage, config: PyPdfToTextConfig | None = None) -> str:
    """
    Generate structured page text from a DocumentPage object in an Azure Document
    Intelligence response.

    Args:
        page: an Azure DocumentPage instance
        config: a PyPdfToTextConfig instance that inherits from the global base config
            `constants` by default. See PyPdfToTextConfig docstring for more info.

    Returns:
        str: page text in a fixed width format that closely adheres to the rendered
            layout in the source pdf.

    """
    config = config or PyPdfToTextConfig()
    if not page.lines:
        return ""
    groups = dedented_groups(
        [CharGroup(_line, page, config) for _line in page.lines or [] if _line.polygon is not None]
    )
    ty_groups = y_coordinate_groups(groups)
    char_width = fixed_char_width(groups)
    lines: list[str] = []
    last_y_coord = 0
    for y_coord, line_data in ty_groups.items():
        if config.PRESERVE_VERTICAL_WHITESPACE and lines:
            fh = line_data[0].effective_height
            blank_lines = (
                0
                if fh == 0
                else (int(abs(y_coord - last_y_coord) / (fh * config.FONT_HEIGHT_WEIGHT)) - 1)
            )
            lines.extend([""] * blank_lines)
        line = ""
        last_disp = 0.0
        for bt_op in line_data:
            offset = int(bt_op.tx // char_width)
            spaces = (offset - len(line)) * (math.ceil(last_disp) < int(bt_op.tx))
            line = f"{line}{' ' * spaces}{bt_op.text}"
            last_disp = bt_op.displaced_tx
        if line.strip() or lines:
            lines.append("".join(c if ord(c) < 14 or ord(c) > 31 else " " for c in line))
        last_y_coord = y_coord
    return "\n".join(
        ln.rstrip() for ln in lines if config.PRESERVE_VERTICAL_WHITESPACE or ln.strip()
    )



"""Manage settings for text extraction and OCR operations"""

from __future__ import annotations

import logging
import os
from dataclasses import dataclass, InitVar, field
from typing import cast, Any, TypedDict


logger = logging.getLogger(__name__)


class PyPdfToTextConfigOverrides(TypedDict, total=False):
    """
    Enumerate settings available to override via the PyPdfToTextConfig
    `overrides` InitVar.
    """

    INHERIT_CONSTANTS: bool
    AZURE_DOCINTEL_ENDPOINT: str
    AZURE_DOCINTEL_SUBSCRIPTION_KEY: str
    AZURE_DOCINTEL_AUTO_CLIENT: bool
    AZURE_DOCINTEL_TIMEOUT: int
    DISABLE_OCR: bool
    DISABLE_PROGRESS_BAR: bool
    PROGRESS_BAR_POSITION: int | None
    FONT_HEIGHT_WEIGHT: float
    OCR_LINE_HEIGHT_SCALE: int
    OCR_POSITIONING_SCALE: int
    PRESERVE_VERTICAL_WHITESPACE: bool
    MAX_CHARS_PER_PDF_PAGE: int
    MIN_LINES_OCR_TRIGGER: int
    TRIGGER_OCR_PAGE_RATIO: float
    SCALE_WEIGHT: float
    MIN_OCR_ROTATION_DEGREES: float
    SUPPRESS_EMBEDDED_TEXT: bool
    REPLACE_BYTE_CODES: dict[bytes, bytes]
    MAX_WORKERS: int
    OCR_HANDWRITTEN_CONFIDENCE_LIMIT: float
    AWS_ACCESS_KEY_ID: str | None
    AWS_SECRET_ACCESS_KEY: str | None
    AWS_SESSION_TOKEN: str | None


@dataclass(kw_only=True)
class _ConfigMixIn:
    """
    Package wide constants for pypdftotext.
    """

    # pylint: disable=invalid-name
    _setattrs_: set[str] = field(default_factory=set, init=False)
    _initialized_: bool = field(default=False, init=False)
    INHERIT_CONSTANTS: bool = True
    """If True (default), values set for the package-wide constants (set via
    pypdftotext.constants.<ATTRIBUTE> = <VALUE>) are inherited by all subsequent
    `PyPdfToTextConfig` instances *by default*."""
    AZURE_DOCINTEL_ENDPOINT: str = field(
        default_factory=lambda: os.getenv("AZURE_DOCINTEL_ENDPOINT", "")
    )
    """The API endpoint of your Azure Document Intelligence instance. Defaults to
    the value of the environment variable of the same name or an empty string."""
    AZURE_DOCINTEL_SUBSCRIPTION_KEY: str = field(
        default_factory=lambda: os.getenv("AZURE_DOCINTEL_SUBSCRIPTION_KEY", "")
    )
    """The API key for your Azure Document Intelligence instance. Defaults to
    the value of the environment variable of the same name or an empty string."""
    AZURE_DOCINTEL_AUTO_CLIENT: bool = True
    """If True (default), the Azure Read OCR client is created automatically
    upon first use."""
    AZURE_DOCINTEL_TIMEOUT: int = 60
    """How long to wait for Azure OCR results before timing out. Default is 60."""
    DISABLE_OCR: bool = False
    """Set to True to disable all OCR operations and return 'code behind' text
    only."""
    DISABLE_PROGRESS_BAR: bool = False
    """Set to True to disable the per page text extraction progress bar (e.g.
    when logging to CloudWatch)."""
    PROGRESS_BAR_POSITION: int | None = None
    """Control position if nesting progress bars from multiple threads."""
    FONT_HEIGHT_WEIGHT: float = 1.0
    """Factor for adjusting line splitting behaviors
    and preserved vertical whitespace in fixed width embedded text output.
    NOTE: Higher values result in fewer blank lines but increase the
    likelihood of triggering a split due to font height based y offsets."""
    OCR_LINE_HEIGHT_SCALE: int = 50
    """Factor between 0 and 100 for adjusting line splitting behaviors
    and preserved vertical whitespace in fixed width OCR text output.
    NOTE: Higher values result in fewer blank lines but increase the
    likelihood of triggering a split due to font height based y offsets."""
    OCR_POSITIONING_SCALE: int = 100
    """The factor by which to upscale the coordinates reported in the
    Azure OCR response when constructing the fixed width layout. Lower
    values result in less spacing and increase the likelihood of combining
    independently reported text fragments onto a single line. Tread with
    caution when messing with this one. Also impacts OCR_LINE_HEIGHT_SCALE
    behavior."""
    PRESERVE_VERTICAL_WHITESPACE: bool = False
    """If False (default), no blank lines will be present in the extracted
    text. If True, blank lines are inserted whenever the nominal font height
    is less than or equal to the y coord displacement."""
    MAX_CHARS_PER_PDF_PAGE: int = 25000
    """The maximum number of characters that can conceivably appear on a single
    PDF page. An 8.5inx11in page packed with nothing 6pt text would contain
    ~17K chars. Some malformed PDFs result in millions of extracted nonsense
    characters which can lead to memory overruns (not to mention bad text).
    If a page contains more characters than this, something is wrong. Clear
    the value and report an empty string."""
    MIN_LINES_OCR_TRIGGER: int = 1
    """A page is marked for OCR if it contains fewer lines in its extracted
    embedded text. OCR only proceeds if a sufficient fraction of the
    total PDF pages have been marked (see `TRIGGER_OCR_PAGE_RATIO`)."""
    TRIGGER_OCR_PAGE_RATIO: float = 0.99
    """OCR will proceed if and only if the fraction of pages with fewer than
    `MIN_LINES_OCR_TRIGGER` lines is greater than this value. Default is 0.99,
    i.e. OCR only occurs if ALL pages hit the minimum lines trigger."""
    SCALE_WEIGHT: float = 1.25
    """Adds priority to contiguously rendered strings when calculating the
    fixed char width."""
    MIN_OCR_ROTATION_DEGREES: float = 1e-5
    """Rotations greater than this value reported by Azure OCR will be applied
    prior to compiling fixed width output."""
    SUPPRESS_EMBEDDED_TEXT: bool = False
    """if true, embedded text extraction will not be attempted. Assuming OCR
    is available, all pages will be OCR'd by default."""
    REPLACE_BYTE_CODES: dict[bytes, bytes] = field(default_factory=dict)
    """A series of byte code substitutions to make in the final extracted text,
    e.g. replacing pdf 'encoded font' checkbox representations with a standard
    unicode ☑ byte sequence."""
    MAX_WORKERS: int = 10
    """The maximum number of threads to initialize during PdfExtractBatch
    parallel operations."""
    OCR_HANDWRITTEN_CONFIDENCE_LIMIT: float = 0.8
    """Azure must be at least this confident that a given span is handwritten
    in order for it to count when determining handwritten character percentage."""
    AWS_ACCESS_KEY_ID: str | None = field(default_factory=lambda: os.getenv("AWS_ACCESS_KEY_ID"))
    """AWS access key ID for the credentials that will be used to pull source
    PDFs from S3 if installed with "s3" extra (`pip install pypdftotext["s3"]`)"""
    AWS_SECRET_ACCESS_KEY: str | None = field(
        default_factory=lambda: os.getenv("AWS_SECRET_ACCESS_KEY")
    )
    """AWS secret access key for the credentials that will be used to pull source
    PDFs from S3 if installed with "s3" extra (`pip install pypdftotext["s3"]`)"""
    AWS_SESSION_TOKEN: str | None = field(default_factory=lambda: os.getenv("AWS_SESSION_TOKEN"))
    """AWS session token for the credentials that will be used to pull source
    PDFs from S3 if installed with "s3" extra (`pip install pypdftotext["s3"]`)"""

    def __post_init__(self):
        self._initialized_ = True

    def __setattr__(self, name: str, value: Any) -> None:
        if self._initialized_ and not name.startswith("_"):  # don't capture internal attrs
            self._setattrs_.add(name)
        return super().__setattr__(name, value)


class _PyPdfToTextConstants(_ConfigMixIn):
    """
    Package wide constants for pypdftotext.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, **kwargs):
        # Only initialize once
        if not hasattr(self, "_initialized_") or not self._initialized_:
            super().__init__(**kwargs)


constants = _PyPdfToTextConstants()


@dataclass(kw_only=True)
class PyPdfToTextConfig(_ConfigMixIn):
    """
    Allows behaviors driven by pypdftotext.constants members to be customized
    on a per PdfExtract basis with minimal hassle.

    Features:
     1. New instances will inherit any setting that has been set programmatically in
        `constants`. Environment variables are re-read on instance creation.
        This allows users to globally set API keys (via env OR `constants`) and other
        desired behaviors (via `constants` only) eliminating the need to supply the
        `config` parameter to every PdfExtract instantiation. I.e. if `config` is not
        defined, a new one that inherits from `constants` / env vars as described
        is created on the fly.
     2. Inheritance from the global constants can be disabled globally by setting
        `constants.INHERIT_CONSTANTS` to False or for a single PyPdfToTextConfig
        instance using the `overrides` parameter:
            `PyPdfToTextConfig(overrides={"INHERIT_CONSTANTS": False})`
        The TypedDict implementation `PdfToTextConfigOverrides` is defined above
        to provide IDE and typing support for the `overrides` parameter.
     3. If a `base` is supplied, its values supersede global constants.
     4. If both a 'base' and 'overrides' are supplied, `overrides` supersede `base`.
     5. You can also manipulate the configs of individual `PdfExtract` and
        `AzureDocIntelIntegrator` instances after creating them:
        ```
        extract = PdfExtract("file.pdf")
        extract.config.<SETTING> = <VALUE>
        # Anything set here ^^ (e.g. an API key) will be applied _before_
        # extraction operations (like OCR).
        for page_text in extract.text_pages:  # triggers extraction/OCR
            ...
        ```

    Examples:
        >>> import os
        >>> from pypdftotext._config import PyPdfToTextConfig, constants

        >>> # Example 1: New instances inherit from constants by default
        >>> constants.DISABLE_OCR = True
        >>> constants.MAX_CHARS_PER_PDF_PAGE = 50000
        >>> config1 = PyPdfToTextConfig()
        >>> config1.DISABLE_OCR
        True
        >>> config1.MAX_CHARS_PER_PDF_PAGE
        50000

        >>> # Example 2: Environment variables are re-read on each instantiation
        >>> os.environ['AZURE_DOCINTEL_ENDPOINT'] = 'https://test.cognitiveservices.azure.com/'
        >>> config2 = PyPdfToTextConfig()
        >>> config2.AZURE_DOCINTEL_ENDPOINT
        'https://test.cognitiveservices.azure.com/'
        >>> os.environ['AZURE_DOCINTEL_ENDPOINT'] = 'https://prod.cognitiveservices.azure.com/'
        >>> config3 = PyPdfToTextConfig()
        >>> config3.AZURE_DOCINTEL_ENDPOINT
        'https://prod.cognitiveservices.azure.com/'
        >>> del os.environ['AZURE_DOCINTEL_ENDPOINT']

        >>> # Example 3: Disable inheritance with INHERIT_CONSTANTS
        >>> constants.DISABLE_PROGRESS_BAR = True
        >>> config4 = PyPdfToTextConfig(overrides={"INHERIT_CONSTANTS": False})
        >>> config4.DISABLE_PROGRESS_BAR  # Not inherited from constants
        False
        >>> config4.INHERIT_CONSTANTS
        False

        >>> # Example 4: Overrides supersede base which supersedes constants
        >>> constants.FONT_HEIGHT_WEIGHT = 1.0
        >>> constants.MIN_OCR_ROTATION_DEGREES = 1e-2
        >>> base_config1 = PyPdfToTextConfig()
        >>> base_config1.MIN_OCR_ROTATION_DEGREES = 0.001
        >>> config5 = PyPdfToTextConfig(
        ...     base=base_config1,
        ...     overrides={"FONT_HEIGHT_WEIGHT": 3.0}
        ... )
        >>> config5.MIN_OCR_ROTATION_DEGREES
        0.001
        >>> config5.FONT_HEIGHT_WEIGHT
        3.0
        >>> # cleanup for doctest:
        >>> constants.DISABLE_OCR = False
        >>> constants.MAX_CHARS_PER_PDF_PAGE = 25000
        >>> constants.MIN_OCR_ROTATION_DEGREES = 1e-5
        >>> constants.DISABLE_PROGRESS_BAR = False
        >>> constants._setattrs_ = set()
    """

    overrides: InitVar[PyPdfToTextConfigOverrides | dict[str, Any] | None] = None
    base: InitVar[PyPdfToTextConfig | None] = None

    def __init__(
        self,
        overrides: PyPdfToTextConfigOverrides | dict[str, Any] | None = None,
        base: PyPdfToTextConfig | None = None,
    ):
        """
        If base is supplied, merge fields by overwriting default values in the
        new instance with non-default values from the base.
        """
        super().__init__()
        super().__post_init__()
        if base or constants.INHERIT_CONSTANTS:
            base = cast(PyPdfToTextConfig, base or constants)
            if base is constants and (overrides or {}).get("INHERIT_CONSTANTS") is False:
                pass  # if overrides disables constant inheritance, don't inherit
            else:
                for field_name in base._setattrs_:  # pylint: disable=protected-access
                    setattr(self, field_name, getattr(base, field_name))
        for field_name, val in (overrides or {}).items():
            if hasattr(self, field_name):
                setattr(self, field_name, val)
            else:
                logger.warning("Ignoring invalid override: %s=%s", field_name, val)
